import { ReplaySubject, of, concat, NEVER, Subscription, Observable, from, BehaviorSubject, fromEvent, Subject } from 'rxjs';
import { materialize, tap, map, distinctUntilChanged, switchMap, shareReplay, filter, withLatestFrom, startWith, publish, takeUntil, mergeAll } from 'rxjs/operators';
import { ɵglobal, Directive, ElementRef, Input, NgModule, ɵmarkDirty, ChangeDetectorRef, TemplateRef, ViewContainerRef, Optional, Pipe } from '@angular/core';

function rxMaterialize() {
    return o$ => o$.pipe(materialize(), tap(({ kind, error }) => {
        if (kind === 'E') {
            console.error(error);
        }
    }), map(({ value, hasValue, error, kind }) => ({
        value,
        hasValue,
        error,
        kind: notificationKindToRxNotificationKind(kind)
    })));
}
function notificationKindToRxNotificationKind(kind) {
    switch (kind) {
        case 'C':
            return 'rxComplete';
        case 'E':
            return 'rxError';
        case 'N':
        default:
            return 'rxNext';
    }
}

/**
 * RenderAware
 *
 * @description
 * This function returns an object that holds all the shared logic for the push pipe and the let directive
 * responsible for change detection
 * If you extend this class you need to implement how the update of the rendered value happens.
 * Also custom behaviour is something you need to implement in the extending class
 */
function createRenderAware(cfg) {
    const strategyName$ = new ReplaySubject(1);
    let currentStrategy;
    const strategy$ = strategyName$.pipe(distinctUntilChanged(), switchMap((stringOrObservable) => typeof stringOrObservable === 'string'
        ? of(stringOrObservable)
        : stringOrObservable), map((strategy) => {
        const s = cfg.strategies[strategy];
        if (!!s) {
            return s;
        }
        throw new Error(`Strategy ${strategy} does not exist.`);
    }), tap((s) => (currentStrategy = s)), 
    // do not repeat the steps before for each subscriber
    shareReplay({ bufferSize: 1, refCount: true }));
    const observablesFromTemplate$ = new ReplaySubject(1);
    const valuesFromTemplate$ = observablesFromTemplate$.pipe(distinctUntilChanged());
    let firstTemplateObservableChange = true;
    const renderingEffect$ = valuesFromTemplate$.pipe(
    // handle null | undefined assignment and new Observable reset
    map((observable$) => {
        if (observable$ === null) {
            return of(null);
        }
        if (!firstTemplateObservableChange) {
            cfg.templateObserver.suspense();
            if (observable$ === undefined) {
                return of(undefined);
            }
        }
        firstTemplateObservableChange = false;
        return observable$;
    }), 
    // forward only observable values
    filter((o$) => o$ !== undefined), distinctUntilChanged(), switchMap((o$) => o$
        // Added behavior will get applied to the observable in `renderWithLatestStrategy`
        .pipe(
    // Forward only distinct values
    distinctUntilChanged(), 
    // Update completion, error and next
    tap(cfg.templateObserver), renderWithLatestStrategy(strategy$))), publish());
    return {
        nextPotentialObservable(value) {
            observablesFromTemplate$.next(value);
        },
        nextStrategy(nextConfig) {
            strategyName$.next(nextConfig);
        },
        rendered$: renderingEffect$,
        activeStrategy$: strategy$,
        subscribe() {
            return new Subscription().add(renderingEffect$.connect());
        }
    };
}
function renderWithLatestStrategy(strategyChanges$) {
    const suspenseNotification = {
        kind: 'rxSuspense',
        value: undefined,
        hasValue: false,
        error: undefined,
    };
    return (o$) => {
        return o$.pipe(rxMaterialize(), withLatestFrom(strategyChanges$), 
        // always use latest strategy on value change
        switchMap(([renderValue, strategy]) => concat(of(renderValue), NEVER).pipe(strategy.rxScheduleCD)), startWith(suspenseNotification));
    };
}

/*
 * createPropertiesWeakMap
 *
 * @param getDefaults: (o: O) => P
 * Example:
 *
 * export interface Properties {
 *   isCoalescing: boolean;
 * }
 *
 * const obj: object = {
 *   foo: 'bar',
 *   isCoalescing: 'weakMap version'
 * };
 *
 * const getDefaults = (ctx: object): Properties => ({isCoalescing: false});
 * const propsMap = createPropertiesWeakMap<object, Properties>(getDefaults);
 *
 * console.log('obj before:', obj);
 * // {foo: "bar", isCoalescing: "weakMap version"}
 * console.log('props before:', propsMap.getProps(obj));
 * // {isCoalescing: "weakMap version"}
 *
 * propsMap.setProps(obj, {isCoalescing: true});
 * console.log('obj after:', obj);
 * // {foo: "bar", isCoalescing: "weakMap version"}
 * console.log('props after:', propsMap.getProps(obj));
 * // {isCoalescing: "true"}
 * */
function createPropertiesWeakMap(getDefaults) {
    const propertyMap = new WeakMap();
    return {
        getProps: getProperties,
        setProps: setProperties
    };
    function getProperties(ctx) {
        const defaults = getDefaults(ctx);
        const propertiesPresent = propertyMap.get(ctx);
        let properties;
        if (propertiesPresent !== undefined) {
            properties = propertiesPresent;
        }
        else {
            properties = {};
            Object.entries(defaults).forEach(([prop, value]) => {
                properties[prop] = hasKey(ctx, prop) ? ctx[prop] : value;
            });
            propertyMap.set(ctx, properties);
        }
        return properties;
    }
    function setProperties(ctx, props) {
        const properties = getProperties(ctx);
        Object.entries(props).forEach(([prop, value]) => {
            properties[prop] = value;
        });
        propertyMap.set(ctx, properties);
        return properties;
    }
    function hasKey(ctx, property) {
        return ctx[property] != null;
    }
}

const coalescingManager = createCoalesceManager();
const ɵ0 = (ctx) => ({
    numCoalescingSubscribers: 0
});
const coalescingContextPropertiesMap = createPropertiesWeakMap(ɵ0);
/**
 * @describe createCoalesceManager
 *
 * returns a
 * Maintains a weak map of component references ans flags
 * them if the coalescing process is already started for them.
 *
 * Used in render aware internally.
 */
function createCoalesceManager() {
    return {
        remove: removeWork,
        add: addWork,
        isCoalescing
    };
    // Increments the number of subscriptions in a scope e.g. a class instance
    function removeWork(scope) {
        const numCoalescingSubscribers = coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers - 1;
        coalescingContextPropertiesMap.setProps(scope, {
            numCoalescingSubscribers: numCoalescingSubscribers >= 0 ? numCoalescingSubscribers : 0
        });
    }
    // Decrements the number of subscriptions in a scope e.g. a class instance
    function addWork(scope) {
        const numCoalescingSubscribers = coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers +
            1;
        coalescingContextPropertiesMap.setProps(scope, {
            numCoalescingSubscribers
        });
    }
    // Checks if anybody else is already coalescing atm
    function isCoalescing(scope) {
        return (coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers >
            0);
    }
}

/**
 * @description
 * Limits the number of synchronous emitted a value from the source Observable to
 * one emitted value per
 *   [`AnimationFrame`](https://developer.mozilla.org/en-US/docs/Web/API/Window/requestAnimationFrame), then repeats
 *   this process for every tick of the browsers event loop.
 *
 * The coalesce operator is based on the [throttle](https://rxjs-dev.firebaseapp.com/api/operators/throttle) operator.
 * In addition to that is provides emitted values for the trailing end only, as well as maintaining a context to scope
 *   coalescing.
 *
 * @param {function(value: T): SubscribableOrPromise} durationSelector - A function
 * that receives a value from the source Observable, for computing the silencing
 * duration for each source value, returned as an Observable or a Promise.
 * It defaults to `requestAnimationFrame` as durationSelector.
 * @param {Object} config - A configuration object to define `leading` and `trailing` behavior and the context object.
 * Defaults to `{ leading: false, trailing: true }`. The default scoping is per subscriber.
 * @return {Observable<T>} An Observable that performs the coalesce operation to
 * limit the rate of emissions from the source.
 *
 * @usageNotes
 * Emit clicks at a rate of at most one click per second
 * ```typescript
 * import { fromEvent, animationFrames } from 'rxjs';
 * import { coalesce } from 'ngRx/component';
 *
 * const clicks = fromEvent(document, 'click');
 * const result = clicks.pipe(coalesce(ev => animationFrames));
 * result.subscribe(x => console.log(x));
 * ```
 */
function coalesceWith(durationSelector, scope) {
    const _scope = scope || {};
    return (source) => {
        const o$ = new Observable((observer) => {
            const rootSubscription = new Subscription();
            rootSubscription.add(source.subscribe(createInnerObserver(observer, rootSubscription)));
            return rootSubscription;
        });
        return o$;
        function createInnerObserver(outerObserver, rootSubscription) {
            let actionSubscription;
            let latestValue;
            const tryEmitLatestValue = () => {
                if (actionSubscription) {
                    // We only decrement the number if it is greater than 0 (isCoalescing)
                    coalescingManager.remove(_scope);
                    if (!coalescingManager.isCoalescing(_scope)) {
                        outerObserver.next(latestValue);
                    }
                }
            };
            return {
                complete: () => {
                    tryEmitLatestValue();
                    outerObserver.complete();
                },
                error: (error) => outerObserver.error(error),
                next: (value) => {
                    latestValue = value;
                    if (!actionSubscription) {
                        // tslint:disable-next-line:no-unused-expression
                        coalescingManager.add(_scope);
                        actionSubscription = durationSelector.subscribe({
                            error: (error) => outerObserver.error(error),
                            next: () => {
                                tryEmitLatestValue();
                                actionSubscription.unsubscribe();
                                actionSubscription = undefined;
                            },
                            complete: () => {
                                tryEmitLatestValue();
                                actionSubscription = undefined;
                            }
                        });
                        rootSubscription.add(new Subscription(() => {
                            tryEmitLatestValue();
                            if (actionSubscription) {
                                actionSubscription.unsubscribe();
                                actionSubscription = undefined;
                            }
                        }));
                    }
                }
            };
        }
    };
}

/**
 * @description
 *
 * A fallback for the new `globalThis` reference.
 *
 *  It should be used to replace `window` due to different environments in:
 *  - SSR (Server Side Rendering)
 *  - Tests
 *  - Browser
 *
 *  @return {globalThis} - A reference to globalThis. `window` in the Browser.
 */
function getGlobalThis() {
    return ɵglobal;
}

/**
 * getZoneUnPatchedApi
 *
 * @description
 *
 * This function returns the zone un-patched API for the a specific Browser API.
 * If no element is passed the window is used instead
 *
 * @param name {string} - The name of the API to check.
 * @param elem {any} - The elem to get un-patched API from.
 * @return {Function} - The zone un-patched API in question.
 *
 */
function getZoneUnPatchedApi(name, elem) {
    elem = elem || getGlobalThis();
    return isApiZonePatched(name, elem) ? elem['__zone_symbol__' + name] : elem[name];
}
/**
 * envZonePatched
 *
 * @description
 *
 * This function checks the window object `zone.js` was instantiated.
 * If so, the `window` object maintains a property named `Zone`.
 *
 * Here how Angular checks it: https://github.com/angular/angular/blob/master/packages/core/src/zone/ng_zone.ts#L123
 *
 * @return {boolean} - true if `zone.js` patched global APIs.
 *
 */
function isEnvZonePatched() {
    return getGlobalThis().Zone !== undefined;
}
/**
 * apiZonePatched
 *
 * @description
 *
 * This function checks if a specific Browser API is patched by `zone.js`.
 *
 * @param name - The name of the API to check.
 * @param elem - The name of the API to check.
 * @return {boolean} - true if `zone.js` patched the API in question.
 *
 */
function isApiZonePatched(name, elem) {
    // if symbol is present, zone patched the API
    return elem['__zone_symbol__' + name] !== undefined;
}
const zoneDetectionCache = new WeakMap();
/**
 * isNgZone
 *
 * @description
 *
 * This function takes an instance of a class which implements the NgZone interface and checks if
 * its `runOutsideAngular()` function calls `apply()` on the function passed as parameter. This
 * means the Angular application that instantiated this service assumes it runs in a ZoneLess
 * environment, and therefore it's change detection will not be triggered by zone related logic.
 *
 * However, keep in mind this does not mean `zone.js` is not present.
 * The environment could still run in ZoneFull mode even if Angular turned it off.
 * Consider the situation of a Angular element configured for ZoneLess
 * environments is used in an Angular application relining on the zone mechanism.
 *
 * @param instance {Class Instance} - The instance to check for constructor name of `NgZone`.
 * @return {boolean} - true if instance is of type `NgZone`.
 *
 */
function isNgZone(instance) {
    const cachedValue = zoneDetectionCache.get(instance);
    if (cachedValue !== undefined) {
        return cachedValue;
    }
    let calledApply = false;
    function fn() {
    }
    fn.apply = () => (calledApply = true);
    instance.runOutsideAngular(fn);
    zoneDetectionCache.set(instance, calledApply);
    return calledApply;
}
/**
 * isNoopNgZone
 *
 *@description
 *
 * This function takes any instance of a class and checks
 * if the constructor name is equal to `NoopNgZone`.
 *
 * For more detailed information read the description of [isNgZone](#isngzone).
 *
 * @param instance {Class Instance} - The instance to check for constructor name of `NoopNgZone`.
 * @return {boolean} - true if instance is of type `NoopNgZone`.
 *
 */
function isNoopNgZone(instance) {
    return !isNgZone(instance);
}

/** A shared promise instance to cause a delay of one microtask */
let resolvedPromise = null;
function getUnpatchedResolvedPromise() {
    resolvedPromise = resolvedPromise || getZoneUnPatchedApi('Promise').resolve();
    return resolvedPromise;
}

function staticCoalesce(work, durationSelector, scope = {}, abC = new AbortController()) {
    let sub;
    if (!coalescingManager.isCoalescing(scope)) {
        coalescingManager.add(scope);
        sub = durationSelector.subscribe(() => {
            tryExecuteWork();
        });
        const abortHandler = function () {
            sub.unsubscribe();
            abC.signal.removeEventListener('abort', abortHandler, false);
        };
        abC.signal.addEventListener('abort', abortHandler, false);
    }
    return abC;
    // =====
    function tryExecuteWork() {
        coalescingManager.remove(scope);
        if (!coalescingManager.isCoalescing(scope)) {
            return work();
        }
        else {
            return void 0;
        }
    }
}

const intervalTick = () => new Observable((subscriber) => {
    const id = window.__zone_symbol__setInterval(() => {
        subscriber.next(0);
        subscriber.complete();
    });
    return () => {
        window.__zone_symbol__clearInterval(id);
    };
});

const timeoutTick = () => new Observable((subscriber) => {
    const id = window.__zone_symbol__setTimeout(() => {
        subscriber.next(0);
        subscriber.complete();
    });
    return () => {
        window.__zone_symbol__clearTimeout(id);
    };
});

const animationFrameTick = () => new Observable((subscriber) => {
    const id = getZoneUnPatchedApi('requestAnimationFrame')(() => {
        subscriber.next(0);
        subscriber.complete();
    });
    return () => {
        getZoneUnPatchedApi('cancelAnimationFrame')(id);
    };
});

// @NOTICE replace logic with 7v handling of promises in RxJS
const promiseTick = () => new Observable((subscriber) => {
    let cancelled = false;
    getUnpatchedResolvedPromise()
        .then(() => {
        if (!cancelled) {
            subscriber.next(0);
            subscriber.complete();
        }
    })
        .catch((e) => {
        subscriber.error(e);
    });
    return () => {
        cancelled = true;
        subscriber.complete();
    };
});

var SchedulingName;
(function (SchedulingName) {
    SchedulingName["animationFrame"] = "animationFrame";
    SchedulingName["Promise"] = "Promise";
    SchedulingName["idleCallback"] = "idleCallback";
    SchedulingName["userBlocking"] = "userBlocking";
    SchedulingName["userVisible"] = "userVisible";
    SchedulingName["background"] = "background";
    SchedulingName["setInterval"] = "setInterval";
})(SchedulingName || (SchedulingName = {}));
var SchedulingPriority;
(function (SchedulingPriority) {
    SchedulingPriority[SchedulingPriority["sync"] = 0] = "sync";
    SchedulingPriority[SchedulingPriority["animationFrame"] = 1] = "animationFrame";
    SchedulingPriority[SchedulingPriority["Promise"] = 2] = "Promise";
    SchedulingPriority[SchedulingPriority["setTimeout"] = 3] = "setTimeout";
    SchedulingPriority[SchedulingPriority["setInterval"] = 4] = "setInterval";
    SchedulingPriority[SchedulingPriority["postMessage"] = 5] = "postMessage";
    SchedulingPriority[SchedulingPriority["idleCallback"] = 6] = "idleCallback";
    SchedulingPriority[SchedulingPriority["userBlocking"] = 7] = "userBlocking";
    SchedulingPriority[SchedulingPriority["userVisible"] = 8] = "userVisible";
    SchedulingPriority[SchedulingPriority["background"] = 9] = "background";
})(SchedulingPriority || (SchedulingPriority = {}));

const priorityTickMap = {
    [SchedulingPriority.animationFrame]: animationFrameTick(),
    [SchedulingPriority.Promise]: promiseTick(),
    [SchedulingPriority.setInterval]: intervalTick(),
    [SchedulingPriority.setTimeout]: timeoutTick(),
};

function staticSchedule(work, priority, abC = new AbortController()) {
    // immediately execute work
    if (priority === false) {
        tryExecuteWork();
        return abC;
    }
    // schedule work
    const sub = priorityTickMap[priority].subscribe({
        // on complete abort further executions
        complete: () => abC.abort(),
        next: () => tryExecuteWork()
    });
    const abortHandler = function () {
        sub.unsubscribe();
        abC.signal.removeEventListener('abort', abortHandler, false);
    };
    abC.signal.addEventListener('abort', abortHandler, false);
    return abC;
    // execute work and abort further executions
    function tryExecuteWork() {
        if (!abC.signal.aborted) {
            work();
            abC.abort();
        }
    }
}

function coalesceAndSchedule(work, priority, scope = {}, abC = new AbortController()) {
    const durationSelector = from(getUnpatchedResolvedPromise());
    const scheduledWork = () => staticSchedule(work, priority, abC);
    const coalesceAbC = staticCoalesce(scheduledWork, durationSelector, scope, abC);
    const abortHandler = function () {
        coalesceAbC.abort();
        abC.signal.removeEventListener('abort', abortHandler, false);
    };
    abC.signal.addEventListener('abort', abortHandler, false);
    return abC;
}

const zonePatchedEvents = [
    'scroll',
    'mousedown',
    'mouseenter',
    'mouseleave',
    'mousemove',
    'mouseout',
    'mouseover',
    'mouseup',
    'load',
    'pointerup',
    'change',
    'blur',
    'focus',
    'click',
    'contextmenu',
    'drag',
    'dragend',
    'dragenter',
    'dragleave',
    'dragover',
    'dragstart',
    'drop',
    'input'
];

/**
 *
 * @description
 *
 * This function takes an elem and event and re-applies the listeners from the passed event to the
 * passed element with the zone un-patched version of it.
 *
 * @param elem {HTMLElement} - The elem to re-apply the listeners to.
 * @param event {string} - The name of the event from which to re-apply the listeners.
 *
 * @returns void
 */
function unpatchEventListener(element, event) {
    // `EventTarget` is patched only in the browser environment, thus
    // running this code on the server-side will throw an exception:
    // `TypeError: element.eventListeners is not a function`.
    if (typeof element.eventListeners !== 'function') {
        return;
    }
    const eventListeners = element.eventListeners(event);
    // Return if no event listeners are present
    if (!eventListeners) {
        return;
    }
    const addEventListener = getZoneUnPatchedApi('addEventListener', element).bind(element);
    eventListeners.forEach((listener) => {
        // Remove and reapply listeners with patched API
        // @TODO use (elem as any).removeAllListeners?(eventName?: string): void;
        element.removeEventListener(event, listener);
        // Reapply listeners with un-patched API
        addEventListener(event, listener);
    });
}
/**
 * @Directive UnpatchEventsDirective
 *
 * @description
 *
 * The `unpatch` directive helps in partially migrating to zone-less apps as well as getting rid
 * of unnecessary renderings through zones `addEventListener` patches.
 * It can be used on any element you apply event bindings.
 *
 * The current way of binding events to the DOM is to use output bindings:
 *  ```html
 * <button (click)="doStuff($event)">click me</button>
 * ```
 *
 * The problem is that every event registered over `()` syntax, e.g. `(click)`
 * marks the component and all its ancestors as dirty and re-renders the whole component tree.
 * This is because zone.js patches the native browser API and whenever one of the patched APIs is used it re-renders.
 *
 * So even if your button is not related to a change that needs a re-render the app will re-render completely.
 * This leads to bad performance. This is especially helpful if you work with frequently fired events like 'mousemove'
 *
 * `unpatch` directive solves that problem.
 *
 * Included Features:
 *  - by default un-patch all registered listeners of the host it is applied on
 *  - un-patch only a specified set of registered event listeners
 *  - works zone independent (it directly checks the widow for patched APIs and un-patches them without the use of `runOutsideZone` which brings more performance)
 *  - Not interfering with any logic executed by the registered callback
 *
 * @usageNotes
 *
 * The `unpatch` directive can be used like shown here:
 * ```html
 * <button [unpatch] (click)="triggerSomeMethod($event)">click me</button>
 * <button [unpatch]="['mousemove']" (mousemove)="doStuff2($event)" (click)="doStuff($event)">click me</button>
 * ```
 *
 * @publicApi
 */
// tslint:disable-next-line:directive-selector
class UnpatchEventsDirective {
    constructor(el) {
        this.el = el;
        this.subscription = new Subscription();
        this.events$ = new BehaviorSubject(zonePatchedEvents);
    }
    reapplyEventListenersZoneUnPatched(events) {
        events.forEach((ev) => {
            unpatchEventListener(this.el.nativeElement, ev);
        });
    }
    ngOnChanges({ events }) {
        if (events && Array.isArray(this.events)) {
            this.events$.next(this.events);
        }
    }
    ngAfterViewInit() {
        this.subscription = this.events$.subscribe((eventList) => {
            this.reapplyEventListenersZoneUnPatched(eventList);
        });
    }
    ngOnDestroy() {
        this.subscription.unsubscribe();
    }
}
UnpatchEventsDirective.decorators = [
    { type: Directive, args: [{ selector: '[unpatch]' },] }
];
UnpatchEventsDirective.ctorParameters = () => [
    { type: ElementRef }
];
UnpatchEventsDirective.propDecorators = {
    events: [{ type: Input, args: ['unpatch',] }]
};

const DECLARATIONS = [UnpatchEventsDirective];
class UnpatchEventsModule {
}
UnpatchEventsModule.decorators = [
    { type: NgModule, args: [{
                declarations: DECLARATIONS,
                exports: DECLARATIONS
            },] }
];

/**
 * TemplateManager
 *
 * @description
 * This function returns an object that holds the logic for managing templates of a `ViewContainerRef`.
 * It abstracts `EmbeddedView` creation, `insert` calls and `ViewContext` updates.
 * Internally it creates template references lazily by combining caching logic and the `ViewContainerRef#detach` method.
 * The `TemplateManager` lets you re-use templates and insert views on-demand, as well as update the view context
 * (e.g. when a new observable notification is sent).
 *
 * @param viewContainerRef reference to a top-level view container where passed templates will be attached
 * @param initialViewContext initial view context state
 */
function createTemplateManager(viewContainerRef, initialViewContext) {
    const templateCache = new Map();
    const viewCache = new Map();
    const viewContext = Object.assign({}, initialViewContext);
    let activeView;
    return {
        hasTemplateRef(name) {
            return templateCache.has(name);
        },
        updateViewContext(viewContextSlice) {
            Object.entries(viewContextSlice).forEach(([key, value]) => {
                viewContext[key] = value;
            });
        },
        addTemplateRef(name, templateRef) {
            assertTemplate(name, templateRef);
            if (!templateCache.has(name)) {
                templateCache.set(name, templateRef);
            }
            else {
                // @Notice We have to think through how this would work. We also call viewCache.set(name, newView); in insertEmbeddedView
                throw new Error('Updating an already existing Template is not supported at the moment.');
            }
        },
        getEmbeddedView(name) {
            return viewCache.get(name);
        },
        displayView(name) {
            if (activeView !== name) {
                if (templateCache.has(name)) {
                    // Detach currently inserted view from the container
                    viewContainerRef.detach();
                    if (viewCache.has(name)) {
                        viewContainerRef.insert(viewCache.get(name));
                    }
                    else {
                        // Creates and inserts view to the view container
                        const newView = viewContainerRef.createEmbeddedView(templateCache.get(name), viewContext);
                        viewCache.set(name, newView);
                    }
                }
                else {
                    // @NOTICE this is here to cause errors and see in which situations we would throw.
                    // In CDK it should work different.
                    console.error(`A non-existing view was tried to insert ${name}`);
                }
                activeView = name;
            }
        },
        destroy() {
            viewCache.forEach((view) => view.destroy());
            viewContainerRef.clear();
        }
    };
}
function assertTemplate(property, templateRef) {
    const isTemplateRefOrNull = !!(!templateRef || templateRef.createEmbeddedView);
    if (!isTemplateRefOrNull) {
        throw new Error(`${property} must be a TemplateRef, but received something else.`);
    }
    return isTemplateRefOrNull;
}

/**
 * @description
 *
 * Noop Strategy
 *
 * The no-operation strategy does nothing.
 * It can be a useful tool for performance improvements as well as debugging
 * The [`[viewport-prio]`](https://github.com/rx-angular/rx-angular/blob/ef99804c1b07aeb96763cacca6afad7bbdab03b1/libs/template/src/lib/experimental/viewport-prio/viewport-prio.directive.ts) directive use it to limit renderings to only visible components:
 *
 * | Name      | Zone Agnostic | Render Method     | Coalescing    | Scheduling |
 * | --------- | --------------| ----------------- | ------------- | ---------- |
 * | `noop`    | ✔             | - `noop`          | ❌             | ❌         |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return {RenderStrategy} - The calculated strategy
 *
 */
function createNoopStrategy(config) {
    return {
        name: 'noop',
        detectChanges: () => { },
        rxScheduleCD: (o) => o.pipe(filter(v => false)),
        scheduleCD: () => new AbortController()
    };
}

/**
 *
 * convenience method to prevent duplicated code. used in strategies after static coalescing
 *
 * @internal
 */
function afterCoalesceAndSchedule(work, afterCD) {
    work();
    if (afterCD) {
        afterCD();
    }
}
/**
 *
 * convenience method to prevent duplicated code.
 * helper for strategies where we have no direct control over the rendering and thus no information about when
 * rendering was finished. used in native & global strategy
 *
 * @internal
 */
function afterScheduleCD(tick) {
    let activeAbortController;
    return function (afterCD) {
        if (activeAbortController) {
            activeAbortController.abort();
        }
        activeAbortController = new AbortController();
        if (afterCD) {
            tick()
                .pipe(takeUntil(fromEvent(activeAbortController.signal, 'abort')))
                .subscribe(() => {
                afterCD();
            });
        }
        return activeAbortController;
    };
}

/**
 * @description
 *
 * Native Strategy
 *
 * @description
 *
 * This strategy mirrors Angular's built-in `async` pipe.
 * This means for every emitted value [`ChangeDetectorRef#markForCheck`](https://github.com/angular/angular/blob/930eeaf177a4c277f437f42314605ff8dc56fc82/packages/core/src/render3/view_ref.ts#L128) is called.
 * Angular still needs zone.js to trigger the [`ApplicationRef#tick`](https://github.com/angular/angular/blob/7d8dce11c0726cdba999fc59a83295d19e5e92e6/packages/core/src/application_ref.ts#L719) to re-render,
 * as the internally called function [`markViewDirty`](https://github.com/angular/angular/blob/930eeaf177a4c277f437f42314605ff8dc56fc82/packages/core/src/render3/instructions/shared.ts#L1837) is only responsible for dirty marking and not rendering.
 *
 * | Name      | Zone Agnostic | Render Method     | Coalescing    | Scheduling               |
 * | --------- | --------------| ----------------- | ------------- | ------------------------ |
 * | `native`  | ❌            | ⮁ `markForCheck` | ✔ RootContext  | `requestAnimationFrame`  |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return {RenderStrategy} - The calculated strategy
 *
 */
function createNativeStrategy(config) {
    const renderMethod = () => config.cdRef.markForCheck();
    const cdScheduled = afterScheduleCD(timeoutTick);
    return {
        name: 'native',
        detectChanges: () => renderMethod(),
        rxScheduleCD: (o) => o.pipe(tap(() => renderMethod()), switchMap(v => timeoutTick().pipe(map(() => v)))),
        scheduleCD: (afterCD) => {
            renderMethod();
            return cdScheduled(afterCD);
        }
    };
}

const promiseDurationSelector = promiseTick();
/**
 * Local Strategies
 *
 * | Name      | Zone Agnostic | Render Method     | Coalescing         | Scheduling                |
 * | --------- | --------------| ----------------- | ------------------ | ------------------------- |
 * | `local`   | ✔             | 🠗 `detectChanges` | ✔ ComponentContext | `requestAnimationFrame`   |
 *
 */
function getLocalStrategies(config) {
    return {
        local: createLocalStrategy(config)
    };
}
/**
 * @description
 *
 * Local Strategy
 *
 * This strategy is rendering the actual component and
 * all it's children that are on a path
 * that is marked as dirty or has components with `ChangeDetectionStrategy.Default`.
 *
 * As detectChanges has no coalescing of render calls
 * like [`ChangeDetectorRef#markForCheck`](https://github.com/angular/angular/blob/930eeaf177a4c277f437f42314605ff8dc56fc82/packages/core/src/render3/view_ref.ts#L128) or [`ɵmarkDirty`](https://github.com/angular/angular/blob/930eeaf177a4c277f437f42314605ff8dc56fc82/packages/core/src/render3/instructions/change_detection.ts#L36) has, so we have to apply our own coalescing, 'scoped' on
 * component level.
 *
 * Coalescing, in this very manner,
 * means **collecting all events** in the same
 * [EventLoop](https://developer.mozilla.org/de/docs/Web/JavaScript/EventLoop) tick, that would cause a re-render and
 * execute **re-rendering only once**.
 *
 * 'Scoped' coalescing, in addition, means **grouping the collected events** by a specific context.
 * E. g. the **component** from which the re-rendering was initiated.
 *
 * This context could be the Component instance or a ViewContextRef,
 * both accessed over the context over `ChangeDetectorRef#context`.
 *
 * | Name      | Zone Agnostic | Render Method     | Coalescing         | Scheduling                 |
 * | --------- | --------------| ----------------- | ------------------ | -------------------------- |
 * | `local`   | ✔             | 🠗 `detectChanges` | ✔ ComponentContext | `requestAnimationFrame`   |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return {RenderStrategy} - The calculated strategy
 *
 */
function createLocalStrategy(config) {
    const component = config.cdRef.context;
    const priority = SchedulingPriority.animationFrame;
    const tick = priorityTickMap[priority];
    const renderMethod = () => {
        config.cdRef.detectChanges();
    };
    const behavior = (o) => o.pipe(coalesceWith(promiseDurationSelector, component), switchMap((v) => tick.pipe(map(() => v))), tap(renderMethod));
    const scheduleCD = (afterCD) => coalesceAndSchedule(() => {
        afterCoalesceAndSchedule(renderMethod, afterCD);
    }, priority, component);
    return {
        name: 'local',
        detectChanges: renderMethod,
        rxScheduleCD: behavior,
        scheduleCD
    };
}

function getGlobalStrategies(config) {
    return {
        global: createGlobalStrategy(config)
    };
}
/**
 *
 * @description
 *
 * Global Strategies
 *
 * This strategy leverages Angular's internal
 *   [`ɵmarkDirty`](https://github.com/angular/angular/blob/930eeaf177a4c277f437f42314605ff8dc56fc82/packages/core/src/render3/instructions/change_detection.ts#L36)
 *   render method. It acts identical to
 *   [`ChangeDetectorRef#markForCheck`](https://github.com/angular/angular/blob/930eeaf177a4c277f437f42314605ff8dc56fc82/packages/core/src/render3/view_ref.ts#L128)
 *   but works also zone-less.
 * `markDirty` in comparison to `markForCheck` also calls
 *   [`scheduleTick`](https://github.com/angular/angular/blob/930eeaf177a4c277f437f42314605ff8dc56fc82/packages/core/src/render3/instructions/shared.ts#L1863)
 *   which is the reason why it also works in zone-less environments.
 *
 * | Name      | Zone Agnostic | Render Method     | Coalescing      | Scheduling       |
 * | --------- | --------------| ----------------- | --------------- | ---------------- |
 * | `global`  | ✔             | ⮁ `ɵmarkDirty`   | ✔ `RootContext` |
 *   [`animationFrame`](https://github.com/angular/angular/blob/930eeaf177a4c277f437f42314605ff8dc56fc82/packages/core/src/render3/util/misc_utils.ts#L39)
 *     |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return { RenderStrategy } - The calculated strategy
 *
 */
function createGlobalStrategy(config) {
    const renderMethod = () => ɵmarkDirty(config.cdRef.context);
    const cdScheduler = afterScheduleCD(animationFrameTick);
    return {
        name: 'global',
        detectChanges: () => renderMethod(),
        rxScheduleCD: (o) => o.pipe(tap(() => renderMethod()), switchMap(v => animationFrameTick().pipe(map(() => v)))),
        scheduleCD: (afterCD) => {
            renderMethod();
            return cdScheduler(afterCD);
        }
    };
}

const promiseDurationSelector$1 = promiseTick();
/**
 * Detach Strategies
 *
 * | Name      | Zone Agnostic | Render Method     | Coalescing         | Scheduling                 |
 * | --------- | --------------| ----------------- | ------------------ | -------------------------- |
 * | `detach`  | ✔             | ⭭ `detectChanges` | ✔ ComponentContext | `requestAnimationFrame`   |
 *
 */
function getDetachStrategies(config) {
    return {
        detach: createDetachStrategy(config)
    };
}
/**
 * Detach Strategy
 *
 * This strategy behaves the same as the local strategy.
 * The detach strategy detaches the component from Angulars change detection.
 * With every new value it re-attaches the component/embedded view to the change detection,
 * renders the new value and detaches again.
 *
 * If a component is detached the input bindings will still receive values.
 * Also the internal logic will work as expected including the use of `ViewChild`.
 * Only the template will not be updated.
 *
 * | Name      | Zone Agnostic | Render Method     | Coalescing         | Scheduling                 |
 * | --------- | --------------| ----------------- | ------------------ | -------------------------- |
 * | `detach`  | ✔             | ⭭ `detectChanges` | ✔ ComponentContext | `requestAnimationFrame`   |
 *
 * @param config - The values this strategy needs to get calculated.
 * @return - The calculated strategy
 *
 */
function createDetachStrategy(config) {
    const component = config.cdRef.context;
    const priority = SchedulingPriority.animationFrame;
    const tick = priorityTickMap[priority];
    const renderMethod = () => {
        config.cdRef.reattach();
        config.cdRef.detectChanges();
        config.cdRef.detach();
    };
    const behavior = (o) => o.pipe(coalesceWith(promiseDurationSelector$1, component), 
    // dispose previous render attempts (tick schedules it into the future)
    switchMap((v) => tick.pipe(map(() => v))), tap(renderMethod));
    const scheduleCD = (afterCD) => coalesceAndSchedule(() => {
        afterCoalesceAndSchedule(renderMethod, afterCD);
    }, priority, component);
    return {
        name: 'detach',
        detectChanges: renderMethod,
        rxScheduleCD: behavior,
        scheduleCD
    };
}

const DEFAULT_STRATEGY_NAME = 'local';
/**
 * @description
 * This method returns the provided strategies as name:strategy pair
 *
 * Built-in Strategies:
 *
 * | Name      | Zone Agnostic | Render Method     | Coalescing         | Scheduling                 |
 * | --------- | --------------| ----------------- | ------------------ | -------------------------- |
 * | `local`   | ✔             | 🠗 `detectChanges` | ✔ ComponentContext | `requestAnimationFrame`   |
 * | `global`  | ✔             | ⮁ `ɵmarkDirty`    | ✔ RootContext     | `requestAnimationFrame`   |
 * | `detach`  | ✔             | ⭭ `detectChanges` | ✔ ComponentContext | `requestAnimationFrame`   |
 * | `noop`    | ✔             | - `noop`          | ❌                 | ❌                        |
 * | `native`  | ❌             | ⮁ `markForCheck` | ✔ RootContext     | `requestAnimationFrame`  |
 *
 * @param config
 */
function getStrategies(config) {
    return Object.assign(Object.assign(Object.assign(Object.assign({}, getLocalStrategies(config)), getGlobalStrategies(config)), getDetachStrategies(config)), { noop: createNoopStrategy(config), native: createNativeStrategy(config) });
}

/**
 * @Directive LetDirective
 *
 * @description
 *
 * The `*rxLet` directive serves a convenient way of binding observables to a view context. Furthermore, it helps
 * you structure view-related models into view context scope (DOM element's scope).
 *
 * Under the hood, it leverages a `RenderStrategy` which in turn takes care of optimizing the change detection
 * of your component or embedded view. The `LetDirective` will render its template and manage change detection after it
 *   got an initial value. So if the incoming `Observable` emits its value lazily (e.g. data coming from `Http`), your
 *   template will be rendered lazily as well. This can very positively impact the initial render performance of your
 *   application.
 *
 *
 * ### Problems with `async` and `*ngIf`
 *
 * In Angular, a way of binding an observable to the view could look like that:
 * ```html
 * <ng-container *ngIf="observableNumber$ | async as n">
 *   <app-number [number]="n"></app-number>
 *   <app-number-special [number]="n"></app-number-special>
 * </ng-container>
 * ```
 *
 * The problem is that `*ngIf` interferes with rendering and in case of a `0` (a falsy value) the component
 * would be hidden. This issue doesn't concern the `LetDirective`.
 *
 * The `AsyncPipe` relies on the Zone to be present - it doesn't really trigger change detection by itself.
 * It marks the component and its children as dirty waiting for the Zone to trigger change detection. So, in case
 * you want to create a zone-less application, the `AsyncPipe` won't work as desired. `LetDirective` comes
 * with its own strategies to manage change detection every time a new notification is sent from
 * the bound Observable.
 *
 *
 * ### Features of `*rxLet`
 *
 * Included features for `*rxLet`:
 * - binding is always present. (see "Problems with `async` and `*ngIf`" section below)
 * - it takes away the multiple usages of the `async` or `push` pipe
 * - a unified/structured way of handling null and undefined
 * - triggers change-detection differently if `zone.js` is present or not (`ChangeDetectorRef.detectChanges` or
 *   `ChangeDetectorRef.markForCheck`)
 * - triggers change-detection differently if ViewEngine or Ivy is present (`ChangeDetectorRef.detectChanges` or
 *   `ɵdetectChanges`)
 * - distinct same values in a row (`distinctUntilChanged` operator),
 * - display custom templates for different observable notifications (rxSuspense, rxNext, rxError, rxComplete)
 * - notify about after changes got rendered to the template (RenderCallback)
 *
 *
 * ### Binding an Observable and using the view context
 *
 * The `*rxLet` directive takes over several things and makes it more convenient and save to work with streams in the
 * template:
 *
 * ```html
 * <ng-container *rxLet="observableNumber$; let n">
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 *
 * <ng-container *rxLet="observableNumber$ as n">
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 * ```
 *
 * In addition to that it provides us information from the whole observable context.
 * We can track the observables:
 * - next value
 * - error occurrence
 * - complete occurrence
 *
 * ```html
 * <ng-container *rxLet="observableNumber$; let n; let e = $rxError, let c = $rxComplete">
 *   <app-number [number]="n" *ngIf="!e && !c"></app-number>
 *   <ng-container *ngIf="e">
 *     There is an error: {{ e }}
 *   </ng-container>
 *   <ng-container *ngIf="c">
 *     Observable completed: {{ c }}
 *   </ng-container>
 * </ng-container>
 * ```
 *
 *
 * ### Using the template-binding
 *
 * You can also use template anchors and display template's content for different observable states:
 * - on complete
 * - on error
 * - on suspense - before the first value is emitted
 *
 * ```html
 * <ng-container
 *   *rxLet="
 *     observableNumber$;
 *     let n;
 *     rxError: error;
 *     rxComplete: complete;
 *     rxSuspense: suspense;
 *   "
 * >
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 * <ng-template #error>ERROR</ng-template>
 * <ng-template #complete>COMPLETE</ng-template>
 * <ng-template #suspense>SUSPENSE</ng-template>
 * ```
 *
 * Internally, `*rxLet` is using a simple "view memoization" - it caches all anchored template references and re-uses
 * them whenever the observable notification (next/error/complete) is sent. Then, it only updates the context
 * (e.g. a value from the observable) in the view.
 *
 *
 * @docsCategory LetDirective
 * @docsPage LetDirective
 * @publicApi
 */
class LetDirective {
    /** @internal */
    constructor(cdRef, nextTemplateRef, viewContainerRef) {
        this.nextTemplateRef = nextTemplateRef;
        this.viewContainerRef = viewContainerRef;
        /** @internal */
        this.subscription = Subscription.EMPTY;
        /** @internal */
        this.initialViewContext = {
            $implicit: undefined,
            rxLet: undefined,
            $rxError: false,
            $rxComplete: false,
            $rxSuspense: false
        };
        /** @internal */
        this.templateObserver = {
            suspense: (value) => {
                this.displayInitialView();
                this.templateManager.updateViewContext({
                    $implicit: value,
                    rxLet: value,
                    $rxError: false,
                    $rxComplete: false,
                    $rxSuspense: true
                });
            },
            next: (value) => {
                this.templateManager.displayView('rxNext');
                this.templateManager.updateViewContext({
                    $implicit: value,
                    rxLet: value
                });
            },
            error: (error) => {
                // fallback to rxNext when there's no template for rxError
                this.templateManager.hasTemplateRef('rxError')
                    ? this.templateManager.displayView('rxError')
                    : this.templateManager.displayView('rxNext');
                this.templateManager.updateViewContext({
                    $rxError: error
                });
            },
            complete: () => {
                // fallback to rxNext when there's no template for rxComplete
                this.templateManager.hasTemplateRef('rxComplete')
                    ? this.templateManager.displayView('rxComplete')
                    : this.templateManager.displayView('rxNext');
                this.templateManager.updateViewContext({
                    $rxComplete: true
                });
            }
        };
        /** @internal */
        this.displayInitialView = () => {
            // Display "suspense" template if provided
            if (this.templateManager.hasTemplateRef('rxSuspense')) {
                this.templateManager.displayView('rxSuspense');
            }
        };
        this.strategies = getStrategies({ cdRef });
        this.templateManager = createTemplateManager(this.viewContainerRef, this.initialViewContext);
        this.renderAware = createRenderAware({
            strategies: this.strategies,
            templateObserver: this.templateObserver
        });
        this.renderAware.nextStrategy(DEFAULT_STRATEGY_NAME);
    }
    /**
     * @description
     * The Observable to be bound to the context of a template.
     *
     * @example
     * <ng-container *rxLet="hero$; let hero">
     *   <app-hero [hero]="hero"></app-hero>
     * </ng-container>
     *
     * @param potentialObservable
     */
    set rxLet(potentialObservable) {
        this.renderAware.nextPotentialObservable(potentialObservable);
    }
    /**
     * @description
     * The rendering strategy to be used when rendering with the reactive context within a template.
     * Use it to dynamically manage your rendering strategy. You can switch the strategies
     * imperatively (with a string) or by bounding an Observable.
     * The default strategy is `'local'`.
     *
     * @example
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *     <ng-container *rxLet="hero$; let hero; strategy: strategy">
     *       <app-hero [hero]="hero"></app-hero>
     *     </ng-container>
     *   `
     * })
     * export class AppComponent {
     *   strategy = 'local';
     * }
     *
     * // OR
     *
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *     <ng-container *rxLet="hero$; let hero; strategy: strategy$">
     *       <app-hero [hero]="hero"></app-hero>
     *     </ng-container>
     *   `
     * })
     * export class AppComponent {
     *   strategy$ = new BehaviorSubject('local');
     * }
     *
     * @param strategy
     * @see {@link strategies}
     */
    set strategy(strategy) {
        this.renderAware.nextStrategy(strategy || DEFAULT_STRATEGY_NAME);
    }
    /**
     * @description
     * A template to show if the bound Observable is in "complete" state.
     *
     * @example
     * <ng-container *rxLet="hero$; let hero; rxComplete: completeTemplate">
     *   <app-hero [hero]="hero"></app-hero>
     * </ng-container>
     * <ng-template #completeTemplate>
     *   <mat-icon>thumb_up</mat-icon>
     * </ng-template>
     *
     * @param templateRef
     */
    set rxComplete(templateRef) {
        this.templateManager.addTemplateRef('rxComplete', templateRef);
    }
    /**
     * @description
     * A template to show if the bound Observable is in "error" state.
     *
     * @example
     * <ng-container *rxLet="hero$; let hero; rxError: errorTemplate">
     *   <app-hero [hero]="hero"></app-hero>
     * </ng-container>
     * <ng-template #errorTemplate>
     *   <mat-icon>thumb_down</mat-icon>
     * </ng-template>
     *
     * @param templateRef
     */
    set rxError(templateRef) {
        this.templateManager.addTemplateRef('rxError', templateRef);
    }
    /**
     * @description
     * A template to show before the first value is emitted from the bound Observable.
     *
     * @example
     * <ng-container *rxLet="hero$; let hero; rxSuspense: suspenseTemplate">
     *   <app-hero [hero]="hero"></app-hero>
     * </ng-container>
     * <ng-template #suspenseTemplate>
     *   <mat-progress-spinner></mat-progress-spinner>
     * </ng-template>
     *
     * @param templateRef
     */
    set rxSuspense(templateRef) {
        this.templateManager.addTemplateRef('rxSuspense', templateRef);
    }
    /** @internal */
    static ngTemplateContextGuard(dir, ctx) {
        return true;
    }
    /** @internal */
    ngOnInit() {
        this.templateManager.addTemplateRef('rxNext', this.nextTemplateRef);
        this.displayInitialView();
        this.subscription = this.renderAware.subscribe();
    }
    /** @internal */
    ngOnDestroy() {
        this.subscription.unsubscribe();
        this.templateManager.destroy();
    }
}
LetDirective.decorators = [
    { type: Directive, args: [{
                selector: '[rxLet]',
                exportAs: 'renderNotifier'
            },] }
];
LetDirective.ctorParameters = () => [
    { type: ChangeDetectorRef },
    { type: TemplateRef },
    { type: ViewContainerRef }
];
LetDirective.propDecorators = {
    rxLet: [{ type: Input }],
    strategy: [{ type: Input, args: ['rxLetStrategy',] }],
    rxComplete: [{ type: Input, args: ['rxLetRxComplete',] }],
    rxError: [{ type: Input, args: ['rxLetRxError',] }],
    rxSuspense: [{ type: Input, args: ['rxLetRxSuspense',] }]
};

/**
 *
 * @description
 *
 * This function takes an elem and event and re-applies the listeners from the passed event to the
 * passed element with the zone un-patched version of it.
 *
 * @param elem {HTMLElement} - The elem to re-apply the listeners to.
 * @param event {string} - The name of the event from which to re-apply the listeners.
 *
 * @returns void
 */
function unpatchEventListener$1(elem, event) {
    const eventListeners = elem.eventListeners(event);
    // Return if no event listeners are present
    if (!eventListeners) {
        return;
    }
    const addEventListener = getZoneUnPatchedApi('addEventListener', elem).bind(elem);
    eventListeners.forEach((listener) => {
        // Remove and reapply listeners with patched API
        elem.removeEventListener(event, listener);
        // Reapply listeners with un-patched API
        addEventListener(event, listener);
    });
}
function intersectionObserver(options) {
    const subject = new Subject();
    const observer = observerSupported()
        ? new IntersectionObserver((entries) => {
            entries.forEach((entry) => subject.next(entry));
        }, options)
        : null;
    const entries$ = new Observable((subscriber) => {
        subject.subscribe(subscriber);
        return () => {
            if (observer) {
                observer.disconnect();
            }
        };
    });
    return {
        entries$,
        observe: observer.observe,
        unobserve: observer.unobserve
    };
}
const observerSupported = () => typeof window !== 'undefined'
    ? !!window.IntersectionObserver
    : false;
const ɵ0$1 = observerSupported;
class ViewportPrioDirective {
    constructor(el, letDirective) {
        this.el = el;
        this.letDirective = letDirective;
        // Note that we're picking only the `intersectionRatio` property
        // since this is the only property that we're intersted in.
        this.entriesSubject = new Subject();
        this.entries$ = this.entriesSubject.pipe(mergeAll());
        this._viewportPrio = 'noop';
        /* @Input('viewport-prio')
        set viewportPrio(prio) {
          if (prio) {
            this._viewportPrio = prio || 'noop';
          }
        }*/
        this.observer = observerSupported()
            ? new IntersectionObserver((entries) => {
                this.entriesSubject.next(entries);
            }, {
                threshold: 0,
            })
            : null;
        this.visibilityEvents$ = this.entries$.pipe(map((entry) => {
            if (entry.intersectionRatio > 0) {
                return 'visible';
            }
            else {
                return 'invisible';
            }
        }));
    }
    ngOnInit() {
        const letStrategyName$ = this.letDirective.renderAware.activeStrategy$.pipe(map((s) => s.name), filter((name) => name !== this._viewportPrio));
        this.visibilityEvents$
            .pipe(withLatestFrom(letStrategyName$), map(([visibility, strategyName]) => visibility === 'visible' ? strategyName : this._viewportPrio))
            .subscribe((strategyName) => {
            this.letDirective.strategy = strategyName;
            // render actual state on viewport enter
            // @TODO this doesnt catch unsubscribe (cant be cancelled)
            this.letDirective.strategies[strategyName].scheduleCD();
        });
        // If the browser doesn't support the `IntersectionObserver` or we're inside
        // the Node.js environment, then this will throw an exception that property
        // `observe` doesn't exist on `null`.
        if (this.observer !== null) {
            this.observer.observe(this.el.nativeElement);
        }
        else {
            // If we're inside the Node.js environment then this should be
            // rendered (e.g. for SEO purposes), and when running this code in browser
            // it will decide itself to render it or not.
            this.entriesSubject.next([{ intersectionRatio: 1 }]);
        }
    }
    ngOnDestroy() {
        if (this.observer) {
            this.observer.disconnect();
        }
    }
}
ViewportPrioDirective.decorators = [
    { type: Directive, args: [{
                // tslint:disable-next-line:directive-selector
                selector: '[viewport-prio]',
            },] }
];
ViewportPrioDirective.ctorParameters = () => [
    { type: ElementRef },
    { type: LetDirective, decorators: [{ type: Optional }] }
];

const DECLARATIONS$1 = [ViewportPrioDirective];
class ViewportPrioModule {
}
ViewportPrioModule.decorators = [
    { type: NgModule, args: [{
                declarations: DECLARATIONS$1,
                exports: DECLARATIONS$1
            },] }
];

/**
 * envRunsIvy
 *
 * @description
 * Determines the used view engine of an Angular project is Ivy or not.
 * The check is done based on following table:
 * | render       | ViewEngine | ViewEngine | Ivy         | Ivy         |
 * | ------------ | ---------- | ---------- | ----------- | ----------- |
 * | **mode**     | prod       | dev        | prod        | dev         |
 * | **ng**       | present    | present    | `undefined` | present     |
 * | **ng.probe** | present    | present    | `undefined` | `undefined` |
 *
 *  So for Ivy we need to make sure that ng is undefined or,
 *  in case of dev environment, ng.probe is undefined.
 *
 * @return {boolean} - true if the used view engine is Ivy.
 *
 */
function isViewEngineIvy() {
    const ng = getGlobalThis().ng;
    // Is the global ng object is unavailable?
    // ng === undefined in Ivy production mode
    // View Engine has the ng object both in development mode and production mode.
    return (ng === undefined ||
        // in case we are in dev mode in ivy
        // `probe` property is available on ng object we use View Engine.
        ng.probe === undefined);
}

/**
 * @Pipe PushPipe
 *
 * @description
 *
 * The `push` pipe serves as a drop-in replacement for the `async` pipe.
 * It contains intelligent handling of change detection to enable us
 * running in zone-full as well as zone-less mode without any changes to the code.
 *
 * The current way of binding an observable to the view looks like that:
 *  ```html
 *  {{observable$ | async}}
 * <ng-container *ngIf="observable$ | async as o">{{o}}</ng-container>
 * <component [value]="observable$ | async"></component>
 * ```
 *
 * The problem is `async` pipe just marks the component and all its ancestors as dirty.
 * It needs zone.js microtask queue to exhaust until `ApplicationRef.tick` is called to render all dirty marked
 *     components.
 *
 * Heavy dynamic and interactive UIs suffer from zones change detection a lot and can
 * lean to bad performance or even unusable applications, but the `async` pipe does not work in zone-less mode.
 *
 * `push` pipe solves that problem.
 *
 * Included Features:
 *  - Take observables or promises, retrieve their values and render the value to the template
 *  - Handling null and undefined values in a clean unified/structured way
 *  - Triggers change-detection differently if `zone.js` is present or not (`detectChanges` or `markForCheck`)
 *  - Distinct same values in a row to increase performance
 *  - Coalescing of change detection calls to boost performance
 *
 * @usageNotes
 *
 * `push` pipe solves that problem. It can be used like shown here:
 * ```html
 * {{observable$ | push}}
 * <ng-container *ngIf="observable$ | push as o">{{o}}</ng-container>
 * <component [value]="observable$ | push"></component>
 * ```
 *
 * @publicApi
 */
class PushPipe {
    constructor(cdRef) {
        this.templateObserver = {
            suspense: () => this.renderedValue = undefined,
            next: (value) => this.renderedValue = value
        };
        this.RenderAware = createRenderAware({
            strategies: getStrategies({
                cdRef
            }),
            templateObserver: this.templateObserver
        });
        this.subscription = this.RenderAware.subscribe();
    }
    transform(potentialObservable, config, renderCallback) {
        const strategy = config || DEFAULT_STRATEGY_NAME;
        this.RenderAware.nextStrategy(strategy);
        this.RenderAware.nextPotentialObservable(potentialObservable);
        return this.renderedValue;
    }
    ngOnDestroy() {
        this.subscription.unsubscribe();
    }
}
PushPipe.decorators = [
    { type: Pipe, args: [{ name: 'push', pure: false },] }
];
PushPipe.ctorParameters = () => [
    { type: ChangeDetectorRef }
];

const DECLARATIONS$2 = [PushPipe];
/**
 * @description
 * This module exports the PushPipe
 *
 * @example
 *
 * ```typescript
 * @NgModule({
 *  imports: [PushModule],
 *  //...
 * })
 * export class AppModule {}
 * ```
 *
 */
class PushModule {
}
PushModule.decorators = [
    { type: NgModule, args: [{
                declarations: DECLARATIONS$2,
                imports: [],
                exports: DECLARATIONS$2
            },] }
];

const EXPORTED_DECLARATIONS = [LetDirective];
class LetModule {
}
LetModule.decorators = [
    { type: NgModule, args: [{
                declarations: EXPORTED_DECLARATIONS,
                exports: [EXPORTED_DECLARATIONS]
            },] }
];

// EXPERIMENTAL

/**
 * Generated bundle index. Do not edit.
 */

export { LetDirective, LetModule, PushModule, PushPipe, SchedulingPriority, UnpatchEventsDirective, UnpatchEventsModule, ViewportPrioDirective, ViewportPrioModule, coalesceWith, coalescingManager, createRenderAware, getStrategies, getZoneUnPatchedApi, isNgZone, isViewEngineIvy, priorityTickMap, renderWithLatestStrategy, rxMaterialize, staticCoalesce };
//# sourceMappingURL=rx-angular-template.js.map
